# This file is part of the markovmodel/mdshare project.
# Copyright (C) 2017-2019 Computational Molecular Biology Group,
# Freie Universitaet Berlin (GER)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import logging
from requests import HTTPError
from hashlib import md5


class LoadError(KeyError):
    def __init__(self, file, message, *args, **kwargs):
        super(LoadError, self).__init__(*args, **kwargs)
        self.file = file
        self.message = message

    def __str__(self):
        return f'{self.file} [{self.message}]'


def file_hash(file, chunk_size=65536):
    """Compute the MD5 hash of a file.

    Arguments:
        file (str): path of the file to be hashed
        chunk_size (int): size of chunks to read
    """
    hash_ = md5()
    with open(file, 'rb') as fh:
        while True:
            data = fh.read(chunk_size)
            if not data:
                break
            hash_.update(data)
    return hash_.hexdigest()


def url_join(repository_url, file):
    """Compose a URL.

    Arguments:
        repository_url (str): url of the repository
        file (str): name of the file in the repository
    """
    return f'{repository_url.rstrip("/")}/{file.lstrip("/")}'


def download_file(repository, file, local_path, callback=None):
    """Download a file.

    Arguments:
        repository (Repository): repository object
        file (str): name of the file in the repository
        local_path (str): local path where the file should be saved
        callback (callable): callback function
    """
    location, metadata = repository.lookup(file)
    logging.debug(
        f'Repository::{location}::{file} has checksum {metadata["hash"]}'
        f' and size {metadata["size"]}')
    logging.debug(
        f'From <{repository.url}> download <{file}> to <{local_path}>')
    response = repository._get_connection().get(
        url_join(repository.url, file),
        stream=True)
    blocksize = 1024 * 8
    with open(local_path, 'wb') as fh:
        for i, data in enumerate(response.iter_content(blocksize)):
            fh.write(data)
            if callback is not None:
                callback(i, blocksize)
    checksum = file_hash(local_path)
    logging.debug(f'Loaded file {local_path} has {checksum}')
    if checksum != metadata['hash']:
        raise LoadError(file, 'checksum test failed')
    return local_path


def attempt_to_download_file(
        repository,
        file,
        local_path,
        max_attempts=3,
        callback=None):
    """Retry to download a file several times if necessary.

    Arguments:
        repository (Repository): repository object
        file (str): name of the file in the repository
        local_path (str): local path where the file should be saved
        max_attempts (int): number of download attempts
        callback (callable): callback function
    """
    attempt = 0
    filename = None

    def fault_handler(filename, exception):
        print(f'error: {exception}', file=sys.stderr)
        try:
            # remove faulty files
            os.unlink(filename)
        except:
            print(f'warning: could not remove file {filename}', file=sys.stderr)
        raise exception

    while attempt < max_attempts:
        attempt += 1
        logging.debug(f'download attempt {attempt}/{max_attempts} ...')
        try:
            filename = download_file(
                repository,
                file,
                local_path,
                callback=callback)
            break
        except (HTTPError, IOError, KeyboardInterrupt) as e:
            fault_handler(filename, e)
    if filename is None:
        raise LoadError(file, 'download failed')
    return filename


def download_wrapper(
        repository, file, working_directory='.',
        max_attempts=3, force=False, callback=None):
    """Download a file if necessary.

    Arguments:
        repository (Repository): repository object
        file (str): name of the file in the repository
        working_directory (str): directory where the file should be saved
        max_attempts (int): number of download attempts
        force (boolean): enforce download even if file exists
        callback (callable): callback function
    """
    logging.debug(
        f'download_wrapper({repository.url}, {file},'
        f' working_directory="{working_directory}",'
        f' max_attempts={max_attempts}, force={force})')
    if working_directory is None:
        raise RuntimeError(
            'working_directory=None is illegal at this point')
    local_path = os.path.join(working_directory, file)
    logging.debug(f'local_path={local_path}')
    if os.path.exists(local_path) and not force:
        logging.debug(f'local_path={local_path} exists ... return')
        return local_path
    logging.debug(
        f'local_path={local_path} does not exist ... attempting download')
    return attempt_to_download_file(
        repository,
        file,
        local_path,
        max_attempts=max_attempts,
        callback=callback)



#!D:\bld\mdshare_1636635969506\_h_env\python.exe

# This file is part of the markovmodel/mdshare project.
# Copyright (C) 2017-2019 Computational Molecular Biology Group,
# Freie Universitaet Berlin (GER)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from mdshare import fetch, Repository
from mdshare.utils import file_hash
from argparse import ArgumentParser
from yaml import load, dump
import fnmatch
import tarfile
import os


def filter_files(files, patterns):
    """Keep only those files which match at least on pattern"""
    include = set()
    for pattern in patterns:
        match = fnmatch.filter(files, pattern)
        include = include | set(match)
    return list(sorted(include))


def get_metadata(file):
    """Get a dict with file hash and size"""
    return dict(
        hash=file_hash(file),
        size=os.path.getsize(file))


def make_container(container, files):
    """Make a .tar.gz container from a list of files"""
    with tarfile.open(container, 'w:gz') as fh:
        for file in files:
            fh.add(file)


def build(template_file):
    """Build the catalogues from the given template"""
    with open(template_file, 'r') as fh:
        template = load(fh)

    for key in ('url', 'include', 'containers'):
        if key not in template:
            raise RuntimeError(f'Cannot build without {key} key')

    db = dict(
        url=template['url'],
        index=dict(),
        containers=dict())

    files = filter_files(os.listdir(), template['include'])
    for file in files:
        db['index'].update({file: get_metadata(file)})

    for container, patterns in template['containers'].items():
        make_container(container, filter_files(files, patterns))
        db['containers'].update({container: get_metadata(container)})

    catalogue = f'{template["name"]}.yaml'
    with open(catalogue, 'w') as fh:
        fh.write(dump(db))

    checksum = f'{template["name"]}.md5'
    with open(checksum, 'w') as fh:
        fh.write(file_hash(catalogue))

    print(f'catalogue written to: {catalogue}')
    print(f'checksum written to:  {checksum}')


def test(catalogue_file, checksum_file):
    repository = Repository(catalogue_file, checksum_file)
    working_directory = 'mdshare-testing-area'
    os.mkdir(working_directory)
    for file in repository.index:
        local_file = fetch(
            file,
            working_directory=working_directory,
            repository=repository)
        os.remove(local_file)
    for file in repository.containers:
        local_files = fetch(
            file,
            working_directory=working_directory,
            repository=repository)
        try:
            os.remove(local_files)
        except TypeError:
            for local_file in local_files:
                os.remove(local_file)
    os.rmdir(working_directory)


if __name__ == '__main__':
    parser = ArgumentParser()
    parser.add_argument(
        'mode',
        help='action to take [ build | test ]',
        metavar='MODE')
    parser.add_argument(
        'yaml',
        help='yaml file with catalogue or catalogue template',
        metavar='FILE')
    parser.add_argument(
        'md5',
        help='md5 checksum file of the catalogue',
        metavar='FILE',
        nargs='?')
    args = parser.parse_args()

    if args.mode.lower() == 'build':
        build(args.yaml)
    elif args.mode.lower() == 'test':
        test(args.yaml, args.md5)
    else:
        raise ValueError(f'Unsupported mode: {args.mode}')



# This file is part of the markovmodel/mdshare project.
# Copyright (C) 2017-2019 Computational Molecular Biology Group,
# Freie Universitaet Berlin (GER)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from humanfriendly import format_size
from yaml import safe_load
import requests
import fnmatch
from .utils import LoadError, file_hash


class Category(dict):
    def __init__(self, data):
        super(Category, self).__init__(data)

    def search(self, pattern):
        return fnmatch.filter(self.keys(), pattern)

    def __str__(self):
        string = ''
        for key in sorted(self.keys()):
            size, unit = format_size(self[key]['size']).split(' ')
            string += f'{key:50s}   {float(size):6.1f} {unit}\n'
        return string.rstrip('\n')


class Repository(object):
    def __init__(self, catalogue_file, checksum_file=None):
        if checksum_file is not None:
            with open(checksum_file, 'r') as fh:
                if file_hash(catalogue_file) != fh.read():
                    raise RuntimeError(
                        'Checksums do not match, check your catalogue files!')
        self.catalogue_file = catalogue_file
        with open(self.catalogue_file, 'r') as fh:
            data = safe_load(fh)
        for key in ('url', 'index', 'containers'):
            if key not in data:
                raise RuntimeError(
                    f'Cannot build repository catalogue without the {key} key')
        self.url = data['url']
        self.index = Category(data['index'])
        self.containers = Category(data['containers'])
        self._connection = None

    def lookup(self, key):
        if key in self.index:
            return 'index', self.index[key]
        elif key in self.containers:
            return 'containers', self.containers[key]
        raise LoadError(key, 'file not in repository catalogue')

    def size(self, key):
        _, data = self.lookup(key)
        return data['size']

    def hash(self, key):
        _, data = self.lookup(key)
        return data['hash']

    def search(self, pattern):
        index = set(self.index.search(pattern))
        containers = set(self.containers.search(pattern))
        return list(sorted(index | containers))

    def stack(self, pattern):
        stack = []
        for file in self.search(pattern):
            location, data = self.lookup(file)
            unpack = location == 'containers'
            stack.append(
                dict(file=file, size=data['size'], unpack=unpack))
        return stack

    def _get_connection(self):
        if self._connection is None:
            self._connection = requests.session()
        return self._connection

    def __str__(self):
        string = f'Repository: {self.url}\n'
        string += f'Files:\n{self.index}\n'
        string += f'Containers:\n{self.containers}'
        return string



# This file is part of the markovmodel/mdshare project.
# Copyright (C) 2017-2019 Computational Molecular Biology Group,
# Freie Universitaet Berlin (GER)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import random
import string
import pytest
import os
from yaml import dump
from ..utils import LoadError
from ..utils import file_hash
from ..repository import Category
from ..repository import Repository


def randomizer(length, pattern=None):
    sample = ''.join([random.choice(string.ascii_letters)
                      for x in range(length)])
    if pattern is None:
        return sample
    return f'{sample}-{pattern}-{sample}'


def make_random_category_dict(n, m):
    def metadata():
        return dict(
            size=random.randint(100, 100000),
            hash=randomizer(32))

    patterns = [randomizer(10) + str(i) for i in range(n)]
    files = []
    for pattern in patterns:
        files += [randomizer(10, pattern) + str(i) for i in range(m)]
    data = {file: metadata() for file in files}
    return patterns, files, data


class RandomCatalogue(object):
    def __init__(self, npattern, nentries, ncontainers, mode=0):
        _, _, index = make_random_category_dict(npattern, nentries)
        _, _, containers = make_random_category_dict(ncontainers, 1)
        self.data = dict(
            url=f'http://{randomizer(10)}.{randomizer(3)}/{randomizer(7)}',
            index=index,
            containers=containers)
        self.offset = ''
        if mode == 1:
            self.data.pop('url')
        elif mode == 2:
            self.data.pop('index')
        elif mode == 3:
            self.data.pop('containers')
        elif mode == 4:
            self.offset = randomizer(42)
        self.file = randomizer(25)

    def __enter__(self):
        with open(f'{self.file}.yaml', 'w') as fh:
            fh.write(dump(self.data))
        with open(f'{self.file}.md5', 'w') as fh:
            fh.write(file_hash(f'{self.file}.yaml') + self.offset)
        return (self.data, self.file)

    def __exit__(self, exception_type, exception_value, traceback):
        os.remove(f'{self.file}.yaml')
        os.remove(f'{self.file}.md5')


def test_category():
    n, m = random.randint(2, 6), random.randint(2, 6)
    patterns, files, data = make_random_category_dict(n, m)
    category = Category(data)
    if len(category.keys()) != n * m:
        raise AssertionError()
    for pattern in patterns:
        if len(category.search(f'*-{pattern}-*')) != m:
            raise AssertionError()
        if len(category.search(f'*-{pattern[1:-1]}-*')) != 0:
            raise AssertionError()
    string = str(category)
    for file in files:
        for key in ('hash', 'size'):
            if category[file][key] != data[file][key]:
                raise AssertionError()
        if file not in string:
            raise AssertionError()


def test_repository():
    args = [random.randint(2, 7) for _ in range(3)]
    with RandomCatalogue(*args, mode=0) as (data, file):
        repository = Repository(f'{file}.yaml', f'{file}.md5')
    string = str(repository)
    if repository.url != data['url']:
        raise AssertionError()
    if data['url'] not in string:
        raise AssertionError()
    for file in data['index']:
        if file not in string:
            raise AssertionError()
        if file not in repository.index:
            raise AssertionError()
        location, metadata = repository.lookup(file)
        if location != 'index':
            raise AssertionError()
        if metadata['size'] != data['index'][file]['size']:
            raise AssertionError()
        if metadata['hash'] != data['index'][file]['hash']:
            raise AssertionError()
        if repository.size(file) != data['index'][file]['size']:
            raise AssertionError()
        if repository.hash(file) != data['index'][file]['hash']:
            raise AssertionError()
    for file in data['containers']:
        if file not in string:
            raise AssertionError()
        if file not in repository.containers:
            raise AssertionError()
        location, metadata = repository.lookup(file)
        if location != 'containers':
            raise AssertionError()
        if metadata['size'] != data['containers'][file]['size']:
            raise AssertionError()
        if metadata['hash'] != data['containers'][file]['hash']:
            raise AssertionError()
        if repository.size(file) != data['containers'][file]['size']:
            raise AssertionError()
        if repository.hash(file) != data['containers'][file]['hash']:
            raise AssertionError()


def test_repository_break():
    for mode in range(4):
        with RandomCatalogue(4, 3, 2, mode=mode + 1) as (data, file):
            with pytest.raises(RuntimeError):
                Repository(f'{file}.yaml', f'{file}.md5')
    args = [random.randint(2, 7) for _ in range(3)]
    with RandomCatalogue(*args, mode=0) as (data, file):
        repository = Repository(f'{file}.yaml', f'{file}.md5')
    for location in ('index', 'containers'):
        for file in data[location]:
            with pytest.raises(LoadError):
                repository.lookup(file[1:-1])



# This file is part of the markovmodel/mdshare project.
# Copyright (C) 2017-2019 Computational Molecular Biology Group,
# Freie Universitaet Berlin (GER)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import os
from ..utils import LoadError
from ..utils import file_hash
from ..api import load_repository
from ..api import search
from ..api import catalogue
from ..api import fetch
from .. import default_repository

FILE = 'mdshare-test-00.txt'
HASH = '5cbb04531c2e9fa7cc1e5d83195a2f81'


def file_check(file):
    if file_hash(file) != HASH:
        raise AssertionError()
    os.remove(file)


def test_load_repository_break():
    with pytest.raises(TypeError):
        load_repository(None)
    with pytest.raises(FileNotFoundError):
        load_repository('not-a-repository')


def test_search():
    if len(search(FILE)) != 1:
        raise AssertionError()
    if search(FILE)[0] != FILE:
        raise AssertionError()
    if len(search(FILE[1:-1])) != 0:
        raise AssertionError()


def test_search_break():
    with pytest.raises(TypeError):
        search(FILE, 'not-a-repository')
    with pytest.raises(TypeError):
        search(None)


def test_catalogue(capsys):
    catalogue()
    captured = capsys.readouterr()
    if captured.out != f'{str(default_repository)}\n':
        raise AssertionError()


def test_catalogue_break():
    with pytest.raises(TypeError):
        catalogue('not-a-repository')


def test_fetch():
    file_check(fetch(FILE))
    file_check(fetch(f'*{FILE[1:-1]}*'))
    file_check(fetch(FILE, repository=default_repository))


def test_fetch_break():
    file = fetch(FILE)
    with pytest.raises(FileExistsError):
        fetch(FILE, working_directory=file)
    os.remove(file)
    with pytest.raises(TypeError):
        fetch(None)
    with pytest.raises(LoadError):
        fetch('not-an-existing-file-or-pattern')
    with pytest.raises(TypeError):
        fetch(FILE, repository='not-a-repository')



#   This file is part of the markovmodel/mdshare project.
#   Copyright (C) 2017-2019 Computational Molecular Biology Group,
#   Freie Universitaet Berlin (GER)
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.


__author__ = 'Christoph Wehmeyer'
__email__ = 'christoph.wehmeyer@fu-berlin.de'
__credits__ = ['Guillermo Pérez-Hernández', 'Martin K. Scherer'],


from pkg_resources import get_distribution, DistributionNotFound
try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    __version__ = 'unknown'
del get_distribution, DistributionNotFound


from .repository import Repository
from os.path import dirname, join
from warnings import warn
try:
    default_repository = Repository(
        join(dirname(__file__), 'data', 'mdshare-catalogue.yaml'),
        join(dirname(__file__), 'data', 'mdshare-catalogue.md5'))
except FileNotFoundError:
    warn('Cannot build the default repository: missing file(s)!')
    default_repository = None
except RuntimeError as e:
    warn(f'Cannot build the default repository: {e.args[0]}')
    default_repository = None
del dirname, join, warn


from .api import load_repository, search, catalogue, fetch
from .utils import LoadError


def load(*args, **kwargs):
    raise NotImplementedError('use fetch')



# This file is part of the markovmodel/mdshare project.
# Copyright (C) 2017-2019 Computational Molecular Biology Group,
# Freie Universitaet Berlin (GER)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import string
import random
import pytest
import os
from .. import default_repository as REPO
from ..utils import LoadError
from ..utils import file_hash
from ..utils import url_join
from ..utils import download_file
from ..utils import attempt_to_download_file
from ..utils import download_wrapper


REPO_URL = REPO.url.rstrip('/')
FILE = 'mdshare-test-00.txt'
HASH = '5cbb04531c2e9fa7cc1e5d83195a2f81'


def local_file():
    file = ''.join([random.choice(string.ascii_letters)
                    for x in range(20)])
    return os.path.join('.', file)


def file_check(file):
    checksum = file_hash(file)
    os.remove(file)
    if checksum != HASH:
        raise AssertionError()


def test_file_hash():
    if file_hash('LICENSE') != 'bb3ca60759f3202f1ae42e3519cd06bc':
        raise AssertionError()


def test_file_hash_break():
    with pytest.raises(TypeError):
        file_hash(None)
    with pytest.raises(FileNotFoundError):
        file_hash('THIS IS NOT A FILE')


def test_url_join():
    url = f'{REPO_URL}/{FILE}'
    if url_join(REPO_URL, FILE) != url:
        raise AssertionError()
    if url_join(REPO_URL, FILE) != url:
        raise AssertionError()
    if url_join(REPO_URL,  f'/{FILE}') != url:
        raise AssertionError()
    if url_join(f'{REPO_URL}/',  f'/{FILE}') != url:
        raise AssertionError()
    if url_join(f'{REPO_URL}//', FILE) != url:
        raise AssertionError()
    if url_join(REPO_URL,  f'//{FILE}') != url:
        raise AssertionError()


def test_url_join_break():
    with pytest.raises(AttributeError):
        url_join(REPO_URL, None)
    with pytest.raises(AttributeError):
        url_join(None, FILE)
    with pytest.raises(AttributeError):
        url_join(REPO_URL, 1)
    with pytest.raises(AttributeError):
        url_join(1, FILE)


def test_download_file():
    file_check(download_file(REPO, FILE, local_file()))


def test_download_file_break():
    with pytest.raises(LoadError):
        download_file(REPO, None, local_file())
    with pytest.raises(LoadError):
        download_file(REPO, 'not-an-existing-file', local_file())
    with pytest.raises(AttributeError):
        download_file(None, FILE, local_file())
    with pytest.raises(AttributeError):
        download_file('not-a-repository', FILE, local_file())


def test_attempt_to_download_file():
    file_check(attempt_to_download_file(REPO, FILE, local_file()))
    file_check(
        attempt_to_download_file(
            REPO, FILE, local_file(), max_attempts=10))


def test_attempt_to_download_file_break():
    with pytest.raises(LoadError):
        attempt_to_download_file(REPO, None, local_file())
    with pytest.raises(AttributeError):
        attempt_to_download_file(None, FILE, local_file())
    with pytest.raises(IsADirectoryError):
        attempt_to_download_file(REPO, FILE, '.')
    with pytest.raises(LoadError):
        attempt_to_download_file(
            REPO, FILE, local_file(), max_attempts=0)
    with pytest.raises(LoadError):
        attempt_to_download_file(
            REPO, 'not-an-existing-file', local_file())


def test_download_wrapper():
    file_check(download_wrapper(REPO, FILE))
    file_check(download_wrapper(REPO, FILE, max_attempts=10))
    with open(FILE, 'w') as fh:
        fh.write('nonsense content')
    if file_hash(download_wrapper(REPO, FILE)) == HASH:
        raise AssertionError()
    file_check(download_wrapper(REPO, FILE, force=True))


def test_download_wrapper_break():
    with pytest.raises(TypeError):
        download_wrapper(REPO, None)
    with pytest.raises(LoadError):
        download_wrapper(REPO, 'not-an-existing-file')
    with pytest.raises(RuntimeError):
        download_wrapper(REPO, FILE, working_directory=None)
    with pytest.raises(AttributeError):
        download_wrapper(None, FILE)
    with pytest.raises(AttributeError):
        download_wrapper('not-a-repository', FILE)
    with pytest.raises(LoadError):
        download_wrapper(REPO, FILE, max_attempts=0)



# This file is part of the markovmodel/mdshare project.
# Copyright (C) 2017-2019 Computational Molecular Biology Group,
# Freie Universitaet Berlin (GER)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import tarfile
from tempfile import mkdtemp
from .utils import LoadError, download_wrapper
from .repository import Repository
from . import default_repository


def load_repository(catalogue_file, checksum_file=None):
    """Load a repository catalogue from file

    Arguments:
        catalogue_file (str): filename of the catalogue
        checksum_file (str): filename of the catalogue's checksum
    """
    return Repository(catalogue_file, checksum_file)


def search(filename_pattern, repository=None):
    """Returns a list of available files matching a filename_pattern.

    Arguments:
        filename_pattern (str): filename pattern, allows for Unix shell-style wildcards
        repository (Repository): repository object
    """
    if repository is None:
        repository = default_repository
    if not isinstance(repository, Repository):
        raise TypeError('received {type(repository)} instead of Repository')
    return repository.search(filename_pattern)


def catalogue(repository=None):
    """Prints a human-friendly list of available files/sizes.

    Arguments:
        repository (Repository): repository object
    """
    if repository is None:
        repository = default_repository
    if not isinstance(repository, Repository):
        raise TypeError('received {type(repository)} instead of Repository')
    print(repository)


def fetch(
        remote_filename, working_directory='.', repository=None,
        max_attempts=3, force=False, show_progress=True):
    """Download a file if it is not already at the traget location.

    Arguments:
        remote_filename (str): name of the file in the repository
        working_directory (str): directory where the file should be saved
        repository (Repository): repository object
        max_attempts (int): number of download attempts
        force (boolean): enforce download even if file exists
        show_progress (boolean): show download progress
    """
    if repository is None:
        repository = default_repository
    if not isinstance(repository, Repository):
        raise TypeError('received {type(repository)} instead of Repository')
    if working_directory is None:
        working_directory = mkdtemp()
    else:
        os.makedirs(working_directory, exist_ok=True)
    try:
        import progress_reporter
        have_progress_reporter = True
    except ImportError:
        have_progress_reporter = False

    stack = repository.stack(remote_filename)
    if len(stack) == 0:
        raise LoadError(remote_filename, 'no match in repository')

    if have_progress_reporter and show_progress:
        callbacks = []
        pg = progress_reporter.ProgressReporter_()
        total = sum(item['size'] for item in stack)

        def update(n, blk, stage):
            downloaded = n * blk
            inc = max(
                0, downloaded - pg._prog_rep_progressbars[stage].n)
            pg.update(inc, stage=stage)
            # total progress
            try:
                pg.update(inc, stage=-1)
            except RuntimeError:
                pass

        from functools import partial
        tqdm_args = dict(unit='B', file=sys.stdout, unit_scale=True)

        n_progress_bars = 0
        for stage, item in enumerate(stack):
            if working_directory is not None:
                path = os.path.join(working_directory, item['file'])
                if os.path.exists(path) and not force:
                    callbacks.append(None)
                else:
                    pg.register(
                        item['size'],
                        description=f'downloading {item["file"]}',
                        tqdm_args=tqdm_args,
                        stage=stage)
                    callbacks.append(partial(update, stage=stage))
                    n_progress_bars += 1
        if n_progress_bars > 1:
            pg.register(
                total, description='total', tqdm_args=tqdm_args, stage=-1)
    else:
        from unittest.mock import MagicMock
        pg = MagicMock()
        callbacks = [None] * len(stack)

    result = []
    with pg.context():
        for item, progress in zip(stack, callbacks):
            file = download_wrapper(
                repository,
                item['file'],
                working_directory=working_directory,
                max_attempts=max_attempts,
                force=force,
                callback=progress)
            if item['unpack']:

                def inspect(members):
                    for member in members:
                        path, filename = os.path.split(member.name)
                        if path == '':
                            yield member, filename

                with tarfile.open(file, 'r:gz') as fh:
                    members = []
                    for member, filename in inspect(fh):
                        members.append(member)
                        result.append(
                            os.path.join(working_directory, filename))
                    fh.extractall(
                        path=working_directory, members=members)
                os.remove(file)
            else:
                result.append(file)

    if len(result) == 0:
        raise LoadError(remote_filename, 'this should not have happend!')
    elif len(result) == 1:
        return result[0]
    return result
